#include <iostream>
#include <fstream>

#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/internet-module.h"
#include "ns3/point-to-point-module.h"
#include "ns3/netanim-module.h"
#include "ns3/applications-module.h"
#include "ns3/point-to-point-layout-module.h"
#include "ns3/traffic-control-helper.h"
#include "ns3/traffic-control-layer.h"

using namespace ns3;

/**********************************/
/*       Various parameters       */
/*                                */
/*  To vary them, move them to    */
/*  the code for cmd-parsing      */
/**********************************/

// 1446 yields packets of 1500 bytes.
// 1448 is the max for default pointtopoint links, yielding packets of size 1502
#define TCPMSS 1446
#define PACKETSIZE (TCPMSS+54)
// this is in segments:
#define TCPINITCWND 3

// this is in packets
#define QLEN 8
// QUEUE_MODE_PACKETS or QUEUE_MODE_BYTES
//#define QMODE "QUEUE_MODE_PACKETS"
#define QMODE "QUEUE_MODE_PACKETS"

// access links have 1Gbps, 1ms
#define BOTTLENECK_CAPACITY "1Mbps"
#define BOTTLENECK_DELAY "8ms"


/**********************************/
/*      MyApp (from fifth.cc)     */
/*                                */
/*  This can be useful to emulate */
/*  app-limited senders (perhaps) */
/**********************************/

class MyApp : public Application
{
public:
    
    MyApp ();
    virtual ~MyApp();
    
    void Setup (Ptr<Socket> socket, Address address, uint32_t packetSize, uint32_t nPackets, DataRate dataRate);
    
private:
    virtual void StartApplication (void);
    virtual void StopApplication (void);
    
    void ScheduleTx (void);
    void SendPacket (void);
    
    Ptr<Socket>     m_socket;
    Address         m_peer;
    uint32_t        m_packetSize;
    uint32_t        m_nPackets;
    DataRate        m_dataRate;
    EventId         m_sendEvent;
    bool            m_running;
    uint32_t        m_packetsSent;
};

MyApp::MyApp ()
: m_socket (0),
m_peer (),
m_packetSize (0),
m_nPackets (0),
m_dataRate (0),
m_sendEvent (),
m_running (false),
m_packetsSent (0)
{
}

MyApp::~MyApp()
{
    m_socket = 0;
}

void
MyApp::Setup (Ptr<Socket> socket, Address address, uint32_t packetSize, uint32_t nPackets, DataRate dataRate)
{
    m_socket = socket;
    m_peer = address;
    m_packetSize = packetSize;
    m_nPackets = nPackets;
    m_dataRate = dataRate;
}

void
MyApp::StartApplication (void)
{
    m_running = true;
    m_packetsSent = 0;
    m_socket->Bind ();
    m_socket->Connect (m_peer);
    SendPacket ();
}

void
MyApp::StopApplication (void)
{
    m_running = false;
    
    if (m_sendEvent.IsRunning ())
    {
        Simulator::Cancel (m_sendEvent);
    }
    
    if (m_socket)
    {
        m_socket->Close ();
    }
}

void
MyApp::SendPacket (void)
{
    Ptr<Packet> packet = Create<Packet> (m_packetSize);
    m_socket->Send (packet);
    
    if (++m_packetsSent < m_nPackets)
    {
        ScheduleTx ();
    }
}

void
MyApp::ScheduleTx (void)
{
    if (m_running)
    {
        Time tNext (Seconds (m_packetSize * 8 / static_cast<double> (m_dataRate.GetBitRate ())));
        m_sendEvent = Simulator::Schedule (tNext, &MyApp::SendPacket, this);
    }
}


/**********************************/
/*      CALLBACKS (for tracing    */
/**********************************/

static void
CwndChange (Ptr<OutputStreamWrapper> stream, uint32_t oldCwnd, uint32_t newCwnd)
{
    // this writes it to stderr...
    //    NS_LOG_UNCOND (Simulator::Now ().GetSeconds () << "\t" << newCwnd);
    
    *stream->GetStream () << Simulator::Now ().GetSeconds () << "\t" << newCwnd <<  "\t" << (newCwnd/TCPMSS) << std::endl;
}

static void
HighSeqChange (Ptr<OutputStreamWrapper> stream, SequenceNumber32 oldValue, SequenceNumber32 newValue)
{
    *stream->GetStream () << Simulator::Now ().GetSeconds () << "\t" << newValue << std::endl;
}

static void
QlenChange (Ptr<OutputStreamWrapper> stream, uint32_t oldQ, uint32_t newQ)
{
    *stream->GetStream () << Simulator::Now ().GetSeconds () << "\t" << newQ << std::endl;
}

/****************************************************/
/****************************************************/

int main (int argc, char *argv[])
{
    
    Config::SetDefault ("ns3::TcpL4Protocol::SocketType", StringValue ("ns3::TcpNewReno"));
    //    Config::SetDefault ("ns3::PointToPointNetDevice::Mtu", UintegerValue (1500));
    
    Config::SetDefault ("ns3::TcpSocket::SegmentSize", UintegerValue (TCPMSS));
    Config::SetDefault ("ns3::TcpSocket::InitialCwnd", UintegerValue (TCPINITCWND));
    
    Config::SetDefault ("ns3::Queue::Mode", StringValue (QMODE));
    Config::SetDefault ("ns3::Queue::MaxPackets", UintegerValue (QLEN));
    Config::SetDefault ("ns3::Queue::MaxBytes", UintegerValue (QLEN*PACKETSIZE));
    
    Config::SetDefault ("ns3::PfifoFastQueueDisc::Limit", UintegerValue (QLEN) );
    
    
    uint32_t    nLeaf = 1; // If non-zero, number of both left and right
    // add more parameters here, e.g. replacing the defines above, to vary them
    CommandLine cmd;
    cmd.AddValue ("nLeaf",     "Number of left and right side leaf nodes", nLeaf);
    cmd.Parse (argc,argv);
    
    // Create the point-to-point link helpers
    PointToPointHelper pointToPointRouter;
    pointToPointRouter.SetDeviceAttribute  ("DataRate", StringValue (BOTTLENECK_CAPACITY));
    pointToPointRouter.SetChannelAttribute ("Delay", StringValue (BOTTLENECK_DELAY));

    // this seems unnecessary
    /*
    pointToPointRouter.SetQueue("ns3::DropTailQueue",
     "Mode", StringValue (QMODE),
     "MaxPackets", UintegerValue (QLEN),
     "MaxBytes", UintegerValue (QLEN*PACKETSIZE));
    */

    // This is how we could set the MTU - but maybe the default is reasonable...
    //
    //    ns3::DropTailQueue<Packet>
    //    Ptr< Queue > q = GetQueue();
    //    pointToPointRouter.SetDeviceAttribute  ("Mtu", UintegerValue (200));

    
    PointToPointHelper pointToPointLeaf;
    pointToPointLeaf.SetDeviceAttribute    ("DataRate", StringValue ("1Gbps"));
    pointToPointLeaf.SetChannelAttribute   ("Delay", StringValue ("1ms"));
    
    PointToPointDumbbellHelper d (nLeaf, pointToPointLeaf,
                                  nLeaf, pointToPointLeaf,
                                  pointToPointRouter);
    
    // Install Stack
    InternetStackHelper stack;
    d.InstallStack (stack);
    
    // Assign IP Addresses
    d.AssignIpv4Addresses (Ipv4AddressHelper ("10.1.1.0", "255.255.255.0"),
                           Ipv4AddressHelper ("10.2.1.0", "255.255.255.0"),
                           Ipv4AddressHelper ("10.3.1.0", "255.255.255.0"));
    // ATTENTION:
    // the IPv4AddressHelper installs a TrafficControl module - how expected!
    // to get a "normal" TCP plot, we need to uninstall this again, else it seems
    // we have an unlimited queue somewhere. Let's de-install it everywhere, and
    // then install only the queue that we really want.
    // https://stackoverflow.com/questions/38951115/tcp-congestion-window-graph-ns-3
    TrafficControlHelper tcHelper;
    tcHelper.Uninstall (d.GetLeft()->GetDevice(0));
    tcHelper.Uninstall (d.GetLeft()->GetDevice(1));
    for (uint32_t i = 0; i < d.LeftCount (); ++i)
    {
        tcHelper.Uninstall (d.GetLeft(i)->GetDevice(0));
        tcHelper.Uninstall (d.GetRight(i)->GetDevice(0));
    }
    // this is how we can install queue disciplines.
    tcHelper.SetRootQueueDisc ("ns3::PfifoFastQueueDisc");
    tcHelper.Install (d.GetLeft()->GetDevice(0));
    // the backwards queue
    //tcHelper.Install (d.GetLeft()->GetDevice(1));
    
    
    Ptr<TrafficControlLayer> tc = d.GetLeft()->GetObject<TrafficControlLayer> ();
    Ptr< QueueDisc > q = tc->GetRootQueueDiscOnDevice (d.GetLeft()->GetDevice(0));

    
    /******** TCP sources and sinks *****/
    
    uint16_t sinkPort = 8080;
    Address sinkAddress (InetSocketAddress (d.GetRightIpv4Address (0), sinkPort));
    PacketSinkHelper packetSinkHelper ("ns3::TcpSocketFactory", InetSocketAddress (Ipv4Address::GetAny (), sinkPort));
    ApplicationContainer sinkApps = packetSinkHelper.Install (d.GetRight (0));
    
    Ptr<Socket> ns3TcpSocket = Socket::CreateSocket (d.GetLeft (0), TcpSocketFactory::GetTypeId ());
    
    Ptr<MyApp> app = CreateObject<MyApp> ();
    // super greedy sender: this puts blocks of 20*MSS size into the send buffer.
    app->Setup (ns3TcpSocket, sinkAddress, TCPMSS*20, 1000000, DataRate ("1Gbps"));
    d.GetLeft (0)->AddApplication (app);
    
    
    /********/
    /*
     // inserting random errors:
     Ptr<RateErrorModel> em = CreateObject<RateErrorModel> ();
     em->SetAttribute ("ErrorRate", DoubleValue (0.00001));
     d.GetRight (0)->GetDevice (0)->SetAttribute ("ReceiveErrorModel", PointerValue (em));
     d.GetLeft (0)->TraceConnectWithoutContext ("PhyRxDrop", MakeCallback (&RxDrop));
     
     */
    
    
    
    /*
     // Background traffic:
     // Install on/off app on all left side nodes except node 0
     OnOffHelper clientHelper ("ns3::UdpSocketFactory", Address ());
     clientHelper.SetAttribute ("OnTime", StringValue ("ns3::UniformRandomVariable"));
     clientHelper.SetAttribute ("OffTime", StringValue ("ns3::UniformRandomVariable"));
     ApplicationContainer clientApps;
     
     for (uint32_t i = 1; i < d.LeftCount (); ++i)
     {
     // Create an on/off app sending packets to the same leaf right side
     AddressValue remoteAddress (InetSocketAddress (d.GetRightIpv4Address (i), 1000));
     clientHelper.SetAttribute ("Remote", remoteAddress);
     clientApps.Add (clientHelper.Install (d.GetLeft (i)));
     }
     clientApps.Start (Seconds (0.0));
     clientApps.Stop (Seconds (10.0));
     */
    
    
    
    // Set up the actual simulation
    Ipv4GlobalRoutingHelper::PopulateRoutingTables ();
    
    //tracing
    AsciiTraceHelper asciiTraceHelper;
    Ptr<OutputStreamWrapper> cwndStream = asciiTraceHelper.CreateFileStream ("cwnd.txt");
    ns3TcpSocket->TraceConnectWithoutContext ("CongestionWindow", MakeBoundCallback (&CwndChange, cwndStream));
    Ptr<OutputStreamWrapper> highSeqStream = asciiTraceHelper.CreateFileStream ("seq.txt");
    ns3TcpSocket->TraceConnectWithoutContext ("HighestSequence", MakeBoundCallback (&HighSeqChange, highSeqStream));
    Ptr<OutputStreamWrapper> qlenStream = asciiTraceHelper.CreateFileStream ("qlen.txt");
    q->TraceConnectWithoutContext ("PacketsInQueue", MakeBoundCallback (&QlenChange, qlenStream));

    //pointToPointRouter.EnablePcapAll ("bottleneck");
    pointToPointLeaf.EnablePcap ("tcpsender", d.GetLeft (0)->GetId (), 0);
    
    // Set the bounding box for animation
    d.BoundingBox (1, 1, 100, 100);
    // Create the animation object and configure for specified output
    AnimationInterface anim ("anim.xml");
    anim.EnablePacketMetadata (); // Optional
    anim.EnableIpv4L3ProtocolCounters (Seconds (0), Seconds (20)); // Optional
    
    
    // schedule events
    // TCP
    app->SetStartTime (Seconds (0.));
    app->SetStopTime (Seconds (10.));
    
    sinkApps.Start (Seconds (0.));
    sinkApps.Stop (Seconds (10.));
    
    Simulator::Stop (Seconds (10));
    Simulator::Run ();
    //    std::cout << "Animation Trace file created:" << animFile.c_str ()<< std::endl;
    Simulator::Destroy ();
    return 0;
}
